/*
 * Copyright (c) 2015-2017 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                           Anthony Mallet on Wed Jun 24 2015
 */
#include <sys/time.h>

#include <ctype.h>
#include <errno.h>
#include <float.h>
#include <fnmatch.h>
#include <math.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

#include "libtk3.h"


/* --- local data ---------------------------------------------------------- */

static struct tk3_channel_s *channels;
static int nchannels;

pthread_mutex_t chanex = PTHREAD_MUTEX_INITIALIZER;
pthread_t chanreader;

static const char *devmagic[] = {
  [TK3_MKBL] = "[?]?mkbl1.8*",
  [TK3_MKFL] = "[?]*[ml][ky][fm][lu]1.8*"
};

static double	get_ts(uint8_t seq, double period, struct timeval tv,
                       time_t origin_ts, struct tk3_data_stamp *s);
static void *	tk3_reader(void *data);


/* --- tk3_client ---------------------------------------------------------- */

int
tk3_client(enum tk3_device dev, const char *path, uint32_t speed,
           char *emsg, size_t emsgsize)
{
  struct tk3_channel_s chan = { .fd = -1 }, *nc;
  struct stat sb;
  int i, s, c;

  chan.fd = tk3_open_tty(path, speed);
  if (chan.fd < 0) goto syserror;

  if (fstat(chan.fd, &sb)) goto syserror;

  /* look for existing channel */
  pthread_mutex_lock(&chanex);
  for(i = 0; i < nchannels; i++) {
    if (channels[i].fd >= 0 &&
        channels[i].st_dev == sb.st_dev &&
        channels[i].st_ino == sb.st_ino) {
      if (dev == channels[i].data.dev)
        channels[i].nrefs++;
      else i = -1;
      pthread_mutex_unlock(&chanex);
      close(chan.fd);
      if (i >= 0) return i;
      snprintf(emsg, emsgsize, "%s: already in use for %s",
               path, (dev == TK3_MKFL) ? "brushless" : "flight control");
      goto error;
    }
  }

  /* test connection */
  while (tk3_recv_msg(&chan, true) == 1); /* flush buffer */
  do {
    c = 0;
    do {
      if (tk3_send_msg(&chan, "?")) /* ask for id */ {
        pthread_mutex_unlock(&chanex);
        goto syserror;
      }

      s = tk3_wait_msg(&(struct tk3_channel_s *){&chan}, 1, &chanex);
      if (s < 0 && errno != EINTR) {
        pthread_mutex_unlock(&chanex);
        goto syserror;
      }
      if (s > 0) break;
    } while(c++ < 3);
    if (c > 3) {
      pthread_mutex_unlock(&chanex);
      errno = ETIMEDOUT;
      goto syserror;
    }

    s = tk3_recv_msg(&chan, true);
  } while(s == 1 && chan.msg[0] != '?');

  chan.msg[chan.len] = 0;
  if (s != 1 || fnmatch(devmagic[dev], (char *)chan.msg, 0)) {
    pthread_mutex_unlock(&chanex);
    if (s != 1) {
      errno = ENOMSG;
      goto syserror;
    } else {
      chan.msg[chan.len] = *"";
      snprintf(emsg, emsgsize,
               "%s: connected device identifies as `%s', expecting `%s'",
               path, chan.msg, devmagic[dev]);
    }
    goto error;
  }

  switch(dev) {
    case TK3_MKBL:
      chan.data.mkbl.fresh = 0;
      chan.data.mkbl.id = chan.msg[1];
      chan.data.mkbl.m_stamp.pseq = 0;
      chan.data.mkbl.m_stamp.pts = 0.;
      chan.data.mkbl.m_stamp.offset = -DBL_MAX;
      break;
    case TK3_MKFL:
      chan.data.mkfl.fresh.imu = 0;
      chan.data.mkfl.fresh.motor = 0;
      chan.data.mkfl.fresh.battery = 0;
      chan.data.mkfl.battery_stamp.pseq = 0;
      chan.data.mkfl.battery_stamp.pts = 0.;
      chan.data.mkfl.battery_stamp.offset = -DBL_MAX;
      chan.data.mkfl.imu_stamp.pseq = 0;
      chan.data.mkfl.imu_stamp.pts = 0.;
      chan.data.mkfl.imu_stamp.offset = -DBL_MAX;
      chan.data.mkfl.motor = NULL;
      chan.data.mkfl.nmotor = 0;
      break;
    case TK3_NONE: break;
  }

  /* create channel */
  nc = realloc(channels, (1 + nchannels) * sizeof(*channels));
  if (!nc) {
    pthread_mutex_unlock(&chanex);
    goto syserror;
  }
  channels = nc;

  if (!nchannels) {
    if (pthread_create(&chanreader, NULL, tk3_reader, NULL)) {
      pthread_mutex_unlock(&chanex);
      goto syserror;
    }
  }

  chan.data.dev = dev;
  strncpy(chan.path, path, sizeof(chan.path));
  chan.st_dev = sb.st_dev;
  chan.st_ino = sb.st_ino;
  chan.nrefs = 1;
  chan.start = 0;
  chan.escape = 0;
  chan.len = 0;
  gettimeofday(&chan.origin_ts, NULL);
  channels[nchannels++] = chan;

  pthread_mutex_unlock(&chanex);
  return nchannels-1;

syserror:
  snprintf(emsg, emsgsize, "%s: %s", path, strerror(errno));
error:
  if (chan.fd >= 0) close(chan.fd);
  return -1;
}

static inline struct tk3_channel_s *
tk3_channel(int c, enum tk3_device dev, char *emsg, size_t emsgsize)
{
  if (c >= 0 && c < nchannels) {
    if (dev == TK3_NONE || channels[c].data.dev == dev)
      return channels + c;

    snprintf(emsg, emsgsize, "%s: bad device", channels[c].path);
    return NULL;
  }

  snprintf(emsg, emsgsize, "%d: bad channel", c);
  return NULL;
}


/* --- tk3_close ----------------------------------------------------------- */

int
tk3_close(int channel, char *emsg, size_t emsgsize)
{
  int i, r;
  struct tk3_channel_s *c = tk3_channel(channel, TK3_NONE, emsg, emsgsize);
  if (!c) return -1;

  if (--c->nrefs) return 0;

  (void)tk3_send_data(channel, "stop", NULL, 0, emsg, emsgsize);
  switch(c->data.dev) {
    case TK3_MKBL:
      (void)tk3_rqst_data(channel, "motor", 0, emsg, emsgsize);
      break;

    case TK3_MKFL:
      (void)tk3_rqst_data(channel, "imu", 0, emsg, emsgsize);
      (void)tk3_rqst_data(channel, "motor", 0, emsg, emsgsize);
      break;

    case TK3_NONE: break;
  }

  pthread_mutex_lock(&chanex);
  if (c->fd >=0) close(c->fd);
  c->fd = -1;

  for(i=r=0; i<nchannels; i++)
    if (channels[i].fd >=0) r++;
  if (r) {
    pthread_mutex_unlock(&chanex);
    return 0;
  }

  nchannels = 0;
  if (channels) free(channels);
  channels = NULL;
  pthread_mutex_unlock(&chanex);

  pthread_join(chanreader, NULL);
  return 0;
}


/* --- tk3_data ------------------------------------------------------------ */

struct tk3_data
tk3_data(int channel, char *emsg, size_t emsgsize)
{
  struct tk3_data d = { .dev = TK3_NONE };
  struct tk3_channel_s *c = tk3_channel(channel, TK3_NONE, emsg, emsgsize);
  if (!c) return d;

  pthread_mutex_lock(&chanex);
  if (c->fd < 0)
    snprintf(emsg, emsgsize, "%s: disconnected", c->path);
  else
    d = c->data;

  switch(c->data.dev) {
    case TK3_MKBL:
      c->data.mkbl.fresh = 0;
      break;

    case TK3_MKFL:
      c->data.mkfl.fresh.imu = 0;
      c->data.mkfl.fresh.motor = 0;
      c->data.mkfl.fresh.battery = 0;
      break;

    case TK3_NONE: break;
  }
  pthread_mutex_unlock(&chanex);
  return d;
}


/* --- tk3_send_data ------------------------------------------------------- */

int
tk3_send_data(int channel, const char *data, double *vals, size_t len,
              char *emsg, size_t emsgsize)
{
  static const struct {
    enum tk3_device dev;
    const char *key, *msg;
  } dmap[] = {
    { TK3_NONE, "start", "g" },
    { TK3_NONE, "stop", "x" },
    { TK3_MKFL, "vel", "w%@" },
    { TK3_MKBL, "vel", "v%@" },
    { TK3_MKFL, "thr", "q%@" },
    { TK3_MKBL, "thr", "p%@" },
    { TK3_NONE, NULL, NULL }
  };

  struct tk3_channel_s *c;
  int16_t d[len];
  int i, n;

  for(i = 0; dmap[i].key; i++) {
    if (!strcmp(dmap[i].key, data)) {
      c = tk3_channel(channel, dmap[i].dev, emsg, emsgsize);
      if (c) break;
    }
  }
  if (!dmap[i].key) {
    snprintf(emsg, emsgsize, "%d: bad request", channel);
    return -1;
  }

  for(n = 0; n < len; n++) {
    if (dmap[i].key[0] == 't')
      d[n] = vals[n] * 1023./100.;
    else
      d[n] = (fabs(vals[n]) < 1000000./65535.) ?
        ((vals[n]<0)?-32767:32767) : 1000000/2/vals[n];
  }

  if (tk3_send_msg(c, dmap[i].msg, d, len)) {
    snprintf(emsg, emsgsize, "%s: %s", c->path, strerror(errno));
    return -1;
  }
  return 0;
}


/* --- tk3_rqst_data ------------------------------------------------------- */

int
tk3_rqst_data(int channel, const char *data, double frequency,
              char *emsg, size_t emsgsize)
{
  static const struct {
    enum tk3_device dev;
    const char *key, *msg;
  } dmap[] = {
    { TK3_MKFL, "battery", "b%4" },
    { TK3_MKFL, "imu", "i%4" },
    { TK3_NONE, "motor", "m%4" },
    { TK3_NONE, NULL, NULL }
  };

  struct tk3_channel_s *c;
  uint32_t p;
  int i;

  for(i = 0; dmap[i].key; i++)
    if (!strcmp(dmap[i].key, data)) break;
  if (!dmap[i].key) {
    snprintf(emsg, emsgsize, "%d: bad request", channel);
    return -1;
  }

  c = tk3_channel(channel, dmap[i].dev, emsg, emsgsize);
  if (!c) return -1;

  p = frequency ? 1000000/frequency : 0;
  if (tk3_send_msg(c, dmap[i].msg, p)) {
    snprintf(emsg, emsgsize, "%s: %s", c->path, strerror(errno));
    return -1;
  }

  switch(i) {
    case 0:
      c->data.mkfl.battery_stamp.pseq = 0;
      c->data.mkfl.battery_stamp.pts = 0.;
      c->data.mkfl.battery_stamp.offset = -DBL_MAX;
      c->data.battery_p = p;
      break;

    case 1:
      c->data.mkfl.imu_stamp.pseq = 0;
      c->data.mkfl.imu_stamp.pts = 0.;
      c->data.mkfl.imu_stamp.offset = -DBL_MAX;
      c->data.imu_p = p;
      break;

    case 2:
      if (c->data.dev == TK3_MKFL)
        c->data.mkfl.nmotor = 0;
      else {
        c->data.mkbl.m_stamp.pseq = 0;
        c->data.mkbl.m_stamp.pts = 0.;
        c->data.mkbl.m_stamp.offset = -DBL_MAX;
      }
      c->data.motor_p = p;
      break;
  }

  return 0;
}


/* --- tk3_gyro_calib ------------------------------------------------------ */

int
tk3_gyro_calib(int channel, int duration, char *emsg, size_t emsgsize)
{
  struct tk3_channel_s *c;

  c = tk3_channel(channel, TK3_MKFL, emsg, emsgsize);
  if (!c) return -1;

  if (tk3_send_msg(c, "zg%1", duration)) {
    snprintf(emsg, emsgsize, "%s: %s", c->path, strerror(errno));
    return -1;
  }

  return 0;
}


/* --- tk3_update ---------------------------------------------------------- */

static void
tk3_update(struct tk3_channel_s *c, const uint8_t *msg, size_t len,
           struct timeval tv)
{
  int16_t v16;
  uint16_t u16;

  switch(*msg++) {
    case 'I': /* IMU data */
      if (len == 14 && c->data.dev == TK3_MKFL) {
        uint8_t seq = *msg++;

        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        c->data.mkfl.acc_x = v16/1000.;

        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        c->data.mkfl.acc_y = v16/1000.;

        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        c->data.mkfl.acc_z = v16/1000.;

        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        c->data.mkfl.gyr_r = v16/1000.;

        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        c->data.mkfl.gyr_p = v16/1000.;

        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        c->data.mkfl.gyr_y = v16/1000.;

        c->data.mkfl.imu_ts = get_ts(
          seq, c->data.imu_p / 1e6, tv, c->origin_ts.tv_sec,
          &c->data.mkfl.imu_stamp);
        c->data.mkfl.fresh.imu = 1;
      }
      break;

    case 'M': /* motor data */
      if (len == 9) {
        struct tk3_mkbl_state *motor;
        uint8_t seq = *msg++;
        uint8_t state = *msg++;
        uint8_t id = state & 0xf;

        switch(c->data.dev) {
          case TK3_MKBL:
            motor = &c->data.mkbl;
            break;

          case TK3_MKFL:
            if (id > c->data.mkfl.nmotor) {
              int i;

              c->data.mkfl.motor =
                realloc(c->data.mkfl.motor, id * sizeof(*c->data.mkfl.motor));
              for(i = c->data.mkfl.nmotor; i < id; i++) {
                c->data.mkfl.motor[i].m_stamp.pts = 0.;
                c->data.mkfl.motor[i].m_stamp.pseq = seq;
                c->data.mkfl.motor[i].m_stamp.offset = -DBL_MAX;
                c->data.mkfl.motor[i].id = i;
                c->data.mkfl.motor[i].velocity = nan("");
                c->data.mkfl.motor[i].pwm = nan("");
                c->data.mkfl.motor[i].current = nan("");
                c->data.mkfl.motor[i].m_ts = nan("");
              }
              c->data.mkfl.nmotor = id;
            }
            motor = &c->data.mkfl.motor[id-1];
            c->data.mkfl.fresh.motor = 1;
            break;

          case TK3_NONE: return;
        }

        motor->id = id;
        motor->emerg = !!(state & 0x80);
        motor->servo = !!(state & 0x40);
        motor->spinning = !!(state & 0x20);
        motor->starting = !!(state & 0x10);

        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        if (motor->spinning)
          motor->velocity = v16 ? 1e6/2/v16 : 0;
        else
          motor->velocity = 0;

        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        motor->pwm = v16 * 100./1024.;

        u16 = ((uint16_t)(*msg++) << 8);
        u16 |= ((uint16_t)(*msg++) << 0);
        motor->current = u16 / 1e3;

        motor->m_ts = get_ts(
          seq, c->data.motor_p / 1e6, tv, c->origin_ts.tv_sec,
          &motor->m_stamp);
        motor->fresh = 1;
      }
      break;

    case 'B': /* battery data */
      if (len == 4 && c->data.dev == TK3_MKFL) {
        uint8_t seq = *msg++;

        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        c->data.mkfl.battery = v16/1000.;

        c->data.mkfl.battery_ts = get_ts(
          seq, c->data.battery_p / 1e6, tv, c->origin_ts.tv_sec,
          &c->data.mkfl.battery_stamp);
        c->data.mkfl.fresh.battery = 1;
      }
      break;
  }
}




/* --- get_ts -------------------------------------------------------------- */

/** Implements Olson, Edwin. "A passive solution to the sensor synchronization
 * problem." International conference on Intelligent Robots and Systems (IROS),
 * 2010 IEEE/RSJ */
static double
get_ts(uint8_t seq, double period, struct timeval tv, time_t origin_ts,
       struct tk3_data_stamp *s)
{
  double ts;
  uint8_t ds;

  /* delta samples */
  ds = seq - s->pseq;
  if (ds > 128)
    /* if too many samples were lost, we might have missed more than 255
     * samples: reset the offset */
    s->offset = -DBL_MAX;
  else
    /* consider a 0.5% clock drift on the sender side */
    s->offset -= 0.005 * ds * period;

  /* update remote timestamp */
  s->pts += ds * period;
  s->pseq = seq;

  /* update offset */
  ts = (tv.tv_sec - origin_ts) + 1e-6 * tv.tv_usec;
  if (s->pts - ts > s->offset)
    s->offset = s->pts - ts;

  /* local timestamp */
  ts = s->pts - s->offset + origin_ts;
  return ts;
}


/* --- tk3_reader ---------------------------------------------------------- */

static void *
tk3_reader(void *data)
{
  struct timeval tv;
  int i, s;

  pthread_mutex_lock(&chanex);
  while(nchannels) {
    s = tk3_wait_msg(&channels, nchannels, &chanex);

    gettimeofday(&tv, NULL);

    for(i = 0; i < nchannels; i++) {
      s = tk3_recv_msg(channels + i, false);
      switch(s) {
        case 1:
          tk3_update(channels + i, channels[i].msg, channels[i].len, tv);
          break;

        case 0:
          break;

        default:
          if (channels[i].fd >=0) close(channels[i].fd);
          channels[i].fd = -1;
      }
    }
  }
  pthread_mutex_unlock(&chanex);

  return NULL;
}
