/*
 * Copyright (c) 2015-2017 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                           Anthony Mallet on Wed Jun 24 2015
 */

#include "mex.h"

#include "libtk3.h"


/* --- local data ---------------------------------------------------------- */

static mxArray *	tk3_mkbl_data(const struct tk3_mkbl_state *data);
static void		tk3_mkbl_data_index(const struct tk3_mkbl_state *data,
				mxArray *r, int n);
static mxArray *	tk3_mkfl_data(const struct tk3_mkfl_state *data);


/* --- tk3.comm ------------------------------------------------------------ */

void
mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  char emsg[256];
  char *key;
  int s;

  if (nrhs < 1)
    mexErrMsgIdAndTxt("tk3:args", "invalid arguments");
  key = mxArrayToString(prhs[0]);
  if (!key)
    mexErrMsgIdAndTxt("tk3:args", "invalid arguments");

  if (!strcmp(key, "connect")) {
    const char *dev, *path;
    double speed;

    if (nrhs != 4)
      mexErrMsgIdAndTxt("tk3:args", "invalid arguments");

    dev = mxArrayToString(prhs[1]);
    path = mxArrayToString(prhs[2]);
    speed = mxGetScalar(prhs[3]);

    if (!dev || !path)
      mexErrMsgIdAndTxt("tk3:args", "invalid arguments");

    if (!strcmp(dev, "mkbl"))
      s = tk3_client(TK3_MKBL, path, speed, emsg, sizeof(emsg));
    else if (!strcmp(dev, "mkfl"))
      s = tk3_client(TK3_MKFL, path, speed, emsg, sizeof(emsg));
    else
      mexErrMsgIdAndTxt("tk3:args", "invalid arguments");

    if (s < 0) mexErrMsgIdAndTxt("tk3:run", emsg);
    mexLock();

    if (nlhs > 0)
      plhs[0] = mxCreateDoubleScalar(s);

    mxFree((char *)path);
    mxFree((char *)dev);

  } else if (!strcmp(key, "disconnect")) {
    int id;

    if (nrhs != 2)
      mexErrMsgIdAndTxt("tk3:args", "invalid arguments");
    id = mxGetScalar(prhs[1]);

    s = tk3_close(id, emsg, sizeof(emsg));
    if (s < 0) mexErrMsgIdAndTxt("tk3:run", emsg);
    mexUnlock();

  } else if (!strcmp(key, "imu") ||
             !strcmp(key, "motor") ||
             !strcmp(key, "battery")) {
    int id;
    double frequency;

    if (nrhs != 3)
      mexErrMsgIdAndTxt("tk3:args", "invalid arguments");
    id = mxGetScalar(prhs[1]);
    frequency = mxGetScalar(prhs[2]);

    s = tk3_rqst_data(id, key, frequency, emsg, sizeof(emsg));
    if (s < 0) mexErrMsgIdAndTxt("tk3:run", emsg);

  } else if (!strcmp(key, "start") ||
             !strcmp(key, "stop") ||
             !strcmp(key, "vels") ||
             !strcmp(key, "vel") ||
             !strcmp(key, "thr")) {
    int id;
    double *vals;
    size_t len;

    if (nrhs != 2 && nrhs != 3)
      mexErrMsgIdAndTxt("tk3:args", "invalid arguments");
    if (nrhs == 3 && !mxIsDouble(prhs[2]))
      mexErrMsgIdAndTxt("tk3:args", "invalid arguments");
    id = mxGetScalar(prhs[1]);
    if (nrhs == 3) {
      vals = mxGetPr(prhs[2]);
      len = mxGetNumberOfElements(prhs[2]);
    } else {
      vals = NULL;
      len = 0;
    }

    s = tk3_send_data(id, key, vals, len, emsg, sizeof(emsg));
    if (s < 0) mexErrMsgIdAndTxt("tk3:run", emsg);

  } else if (!strcmp(key, "data")) {
    int id;
    struct tk3_data d;

    if (nlhs < 1) return;
    if (nrhs != 2)
      mexErrMsgIdAndTxt("tk3:args", "invalid arguments");
    id = mxGetScalar(prhs[1]);

    d = tk3_data(id, emsg, sizeof(emsg));
    if (d.dev == TK3_NONE)
      mexErrMsgIdAndTxt("tk3:run", emsg);

    switch(d.dev) {
      case TK3_MKBL:
        plhs[0] = tk3_mkbl_data(&d.mkbl);
        break;

      case TK3_MKFL:
        plhs[0] = tk3_mkfl_data(&d.mkfl);
        break;

      case TK3_NONE: break;
    }

  } else if (!strcmp(key, "gcalib")) {
    int id;
    double duration;

    if (nrhs != 3)
      mexErrMsgIdAndTxt("tk3:args", "invalid arguments");
    id = mxGetScalar(prhs[1]);
    duration = mxGetScalar(prhs[2]);

    s = tk3_gyro_calib(id, duration, emsg, sizeof(emsg));
    if (s < 0) mexErrMsgIdAndTxt("tk3:run", emsg);

  } else
    mexErrMsgIdAndTxt("genomix:args", "invalid command");

  mxFree(key);
}


/* --- tk3_mkbl_data ------------------------------------------------------- */

static const char *mkbl_fields[] = {
  "id",
  "emerg", "servo", "spinning", "starting",
  "velocity", "pwm", "current",
  "m_ts"
};

static mxArray *
tk3_mkbl_data(const struct tk3_mkbl_state *data)
{
  mxArray *r;

  r = mxCreateStructArray(
    1, &(mwSize){1}, sizeof(mkbl_fields)/sizeof(mkbl_fields[0]), mkbl_fields);
  if (!r) return NULL;

  if (data->fresh)
    tk3_mkbl_data_index(data, r, 0);

  return r;
}

static void
tk3_mkbl_data_index(const struct tk3_mkbl_state *data, mxArray *r, int n)
{
  mxArray *v;
  int i;

  i = 0;
  v = mxCreateDoubleScalar(data->id);
  mxSetFieldByNumber(r, n, i++, v);
  v = mxCreateDoubleScalar(data->emerg);
  mxSetFieldByNumber(r, n, i++, v);
  v = mxCreateDoubleScalar(data->servo);
  mxSetFieldByNumber(r, n, i++, v);
  v = mxCreateDoubleScalar(data->spinning);
  mxSetFieldByNumber(r, n, i++, v);
  v = mxCreateDoubleScalar(data->starting);
  mxSetFieldByNumber(r, n, i++, v);
  v = mxCreateDoubleScalar(data->velocity);
  mxSetFieldByNumber(r, n, i++, v);
  v = mxCreateDoubleScalar(data->pwm);
  mxSetFieldByNumber(r, n, i++, v);
  v = mxCreateDoubleScalar(data->current);
  mxSetFieldByNumber(r, n, i++, v);
  v = mxCreateDoubleScalar(data->m_ts);
  mxSetFieldByNumber(r, n, i++, v);
}


/* --- tk3_mkfl_data ------------------------------------------------------- */

static const char *mkfl_fields[] = {
  "battery", "battery_ts",
  "acc_x", "acc_y", "acc_z", "gyr_r", "gyr_p", "gyr_y", "imu_ts",
  "motor"
};

static mxArray *
tk3_mkfl_data(const struct tk3_mkfl_state *data)
{
  mxArray *r, *v, *m;
  int i, n;

  r = mxCreateStructArray(
    1, &(mwSize){1}, sizeof(mkfl_fields)/sizeof(mkfl_fields[0]), mkfl_fields);
  if (!r) return NULL;

  i = 0;
  if (data->fresh.battery) {
    v = mxCreateDoubleScalar(data->battery);
    mxSetFieldByNumber(r, 0, i++, v);
    v = mxCreateDoubleScalar(data->battery_ts);
    mxSetFieldByNumber(r, 0, i++, v);
  } else
    i+=2;

  if (data->fresh.imu) {
    v = mxCreateDoubleScalar(data->acc_x);
    mxSetFieldByNumber(r, 0, i++, v);
    v = mxCreateDoubleScalar(data->acc_y);
    mxSetFieldByNumber(r, 0, i++, v);
    v = mxCreateDoubleScalar(data->acc_z);
    mxSetFieldByNumber(r, 0, i++, v);

    v = mxCreateDoubleScalar(data->gyr_r);
    mxSetFieldByNumber(r, 0, i++, v);
    v = mxCreateDoubleScalar(data->gyr_p);
    mxSetFieldByNumber(r, 0, i++, v);
    v = mxCreateDoubleScalar(data->gyr_y);
    mxSetFieldByNumber(r, 0, i++, v);

    v = mxCreateDoubleScalar(data->imu_ts);
    mxSetFieldByNumber(r, 0, i++, v);
  } else
    i+=7;

  if (data->fresh.motor) {
    m = mxCreateStructArray(
      1, &(mwSize){data->nmotor},
      sizeof(mkbl_fields)/sizeof(mkbl_fields[0]), mkbl_fields);
    if (!m) return NULL;
    for(n = 0; n < data->nmotor; n++) {
      tk3_mkbl_data_index(&data->motor[n], m, n);
    }
    mxSetFieldByNumber(r, 0, i++, m);
  }

  return r;
}
